import { assertEquals } from "https://deno.land/std@0.224.0/assert/mod.ts"

export class Ingredient {
	public constructor(
		public name: string,
		public amount: number,
	) {}
}

export class Recipe {
	public constructor(
		public name: string,
		public ingredients: Ingredient[],
	) {}
}

export class RecipeSet {
	public constructor(
		public recipes = new Map<string, Recipe>(),
	) {}

	public append(recipe: Recipe) {
		this.recipes.set(recipe.name, recipe)
	}

	public static parse(text: string) {
		const main = new Recipe("main", [])
		const result = new RecipeSet()
		let current_block = null as Recipe | null
		function append_current() {
			if (current_block !== null) result.append(current_block)
			current_block = null
		}
		for (const line of text.split("\n")) {
			if (line.trim().startsWith("#")) continue
			if (line.trim() === "") {
				append_current()
				continue
			}
			const [count_str, ...words] = line.split(" ")
			const count = parseFloat(count_str)
			if (isNaN(count)) {
				append_current()
				current_block = new Recipe(line, [])
			} else {
				const ingredient = new Ingredient(words.join(" "), count)
				if (current_block === null) main.ingredients.push(ingredient)
				else current_block.ingredients.push(ingredient)
			}
		}
		append_current()
		return [main, result] as const
	}
}

Deno.test("test_recipe_set_parse", () => {
	const [main, recipe_set] = RecipeSet.parse(`
a
1 b

b
2 c

2 a
1 c
`)
	assertEquals(
		recipe_set,
		new RecipeSet(
			new Map([
				["a", new Recipe("a", [new Ingredient("b", 1)])],
				["b", new Recipe("b", [new Ingredient("c", 2)])],
			]),
		),
	)

	assertEquals(main, new Recipe("main", [new Ingredient("a", 2), new Ingredient("c", 1)]))
})

export class Step {
	public constructor(
		public name: string,
		public amount: number,
		public deps: Step[],
	) {}

	public static resolve(root: Recipe, recipes: RecipeSet): Step | CircDepErr {
		function recurse(recipe: Recipe, multiplier: number, path: string[]): Step | CircDepErr {
			const next_parents = [...path, recipe.name]
			const steps = [] as Step[]
			for (const ingr of recipe.ingredients) {
				if (next_parents.includes(ingr.name)) return new CircDepErr(next_parents)
				const ingr_multiplier = ingr.amount * multiplier
				const ingr_recipe = recipes.recipes.get(ingr.name)
				if (ingr_recipe === undefined) {
					steps.push(new Step(ingr.name, ingr_multiplier, []))
					continue
				}
				const res = recurse(ingr_recipe, ingr_multiplier, next_parents)
				if (res instanceof CircDepErr) return res
				steps.push(res)
			}
			return new Step(recipe.name, multiplier, steps)
		}
		return recurse(root, 1, [])
	}

	public display() {
		function* rec(step: Step): Generator<string> {
			yield `${step.amount} ${step.name}`
			const deps = [...step.deps]
			if (deps.length < 1) return
			const [last] = deps.splice(deps.length - 1, 1)
			for (const dep of deps) {
				const [first, ...lines] = rec(dep)
				yield `├─${first}`
				for (const line of lines) yield `│ ${line}`
			}
			const [first, ...lines] = rec(last)
			yield `└─${first}`
			for (const line of lines) yield `  ${line}`
		}
		return [...this.deps.map(rec).map((l) => l.toArray())].flat().join("\n")
	}

	public intermediaries() {
		function* rec(step: Step): Generator<[string, number]> {
			if (step.deps.length > 0) yield [step.name, step.amount]
			for (const dep of step.deps) yield* rec(dep)
		}
		const result = new DefaultMap((_: string) => ({ total: 0 }))
		for (const dep of this.deps) for (const [name, value] of rec(dep)) result.get(name).total += value
		return result.map.entries()
			.map(([k, { total }]) => [k, total] as const)
			.toArray()
			.toSorted(([a], [b]) => a.localeCompare(b))
	}

	public inputs() {
		function* rec(step: Step): Generator<[string, number]> {
			if (step.deps.length < 1) yield [step.name, step.amount]
			for (const dep of step.deps) yield* rec(dep)
		}
		const result = new DefaultMap((_: string) => ({ total: 0 }))
		for (const dep of this.deps) for (const [name, value] of rec(dep)) result.get(name).total += value
		return result.map.entries()
			.map(([k, { total }]) => [k, total] as const)
			.toArray()
			.toSorted(([a], [b]) => a.localeCompare(b))
	}
}

export class CircDepErr {
	public constructor(
		public path: string[],
	) {}
}

Deno.test("test_display_lines", () => {
	const [main, recipe_set] = RecipeSet.parse(`
a
1 b
1 c

b
2 c

2 a
1 c
	`.trim())
	const root = Step.resolve(main, recipe_set)
	if (root instanceof CircDepErr) throw new Error()
	const expects = `
1 main
├─2 a
│ ├─2 b
│ │ └─4 c
│ └─2 c
└─1 c
	`.trim()
	assertEquals(root.display(), expects)
})

export class DefaultMap<K, V> {
	public map = new Map<K, V>()

	constructor(
		public default_construct: (key: K) => V,
	) {}

	get(key: K) {
		if (!this.map.has(key)) this.map.set(key, this.default_construct(key))
		return this.map.get(key) as V
	}
}
